<?php

declare(strict_types = 1);

namespace Drupal\entity_share_client\Service;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\EntityReferenceFieldItemListInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Language\LanguageManager;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\StreamWrapper\StreamWrapperManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\entity_share_client\Entity\RemoteInterface;
use Drupal\entity_share_client\Event\RelationshipFieldValueEvent;
use Drupal\file\FileInterface;
use Drupal\jsonapi\Normalizer\JsonApiDocumentTopLevelNormalizer;
use Drupal\jsonapi\ResourceType\ResourceTypeRepository;
use GuzzleHttp\Exception\ClientException;
use Psr\Log\LoggerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\Serializer\SerializerInterface;

/**
 * Class JsonapiHelper.
 *
 * @package Drupal\entity_share_client\Service
 */
class JsonapiHelper implements JsonapiHelperInterface {
  use StringTranslationTrait;

  /**
   * The JsonApiDocumentTopLevelNormalizer normalizer.
   *
   * @var \Drupal\jsonapi\Normalizer\JsonApiDocumentTopLevelNormalizer
   */
  protected $jsonapiDocumentTopLevelNormalizer;

  /**
   * The resource type repository.
   *
   * @var \Drupal\jsonapi\ResourceType\ResourceTypeRepository
   */
  protected $resourceTypeRepository;

  /**
   * The bundle infos from the website.
   *
   * @var array
   */
  protected $bundleInfos;

  /**
   * The entity type definitions.
   *
   * @var \Drupal\Core\Entity\EntityTypeInterface[]
   */
  protected $entityDefinitions;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The stream wrapper manager.
   *
   * @var \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface
   */
  protected $streamWrapperManager;

  /**
   * The language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * The remote manager.
   *
   * @var \Drupal\entity_share_client\Service\RemoteManagerInterface
   */
  protected $remoteManager;

  /**
   * The event dispatcher service.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The request service.
   *
   * @var \Drupal\entity_share_client\Service\RequestServiceInterface
   */
  protected $requestService;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * A prepared HTTP client for file transfer.
   *
   * @var \GuzzleHttp\Client
   */
  protected $fileHttpClient;

  /**
   * A prepared HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The remote website on which to prepare the clients.
   *
   * @var \Drupal\entity_share_client\Entity\RemoteInterface
   */
  protected $remote;

  /**
   * Logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The state information service.
   *
   * @var \Drupal\entity_share_client\Service\StateInformationInterface
   */
  protected $stateInformation;

  /**
   * The list of the currently imported entities.
   *
   * @var array
   */
  protected $importedEntities;

  /**
   * JsonapiHelper constructor.
   *
   * @param \Symfony\Component\Serializer\SerializerInterface $serializer
   *   A serializer.
   * @param \Drupal\jsonapi\Normalizer\JsonApiDocumentTopLevelNormalizer $jsonapi_document_top_level_normalizer
   *   The JsonApiDocumentTopLevelNormalizer normalizer.
   * @param \Drupal\jsonapi\ResourceType\ResourceTypeRepository $resource_type_repository
   *   The resource type repository.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The entity type bundle info service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface $stream_wrapper_manager
   *   The stream wrapper manager.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager.
   * @param \Drupal\entity_share_client\Service\RemoteManagerInterface $remote_manager
   *   The remote manager service.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher service.
   * @param \Drupal\entity_share_client\Service\RequestServiceInterface $request_service
   *   The request service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger service.
   * @param \Drupal\entity_share_client\Service\StateInformationInterface $state_information
   *   The state information service.
   */
  public function __construct(
    SerializerInterface $serializer,
    JsonApiDocumentTopLevelNormalizer $jsonapi_document_top_level_normalizer,
    ResourceTypeRepository $resource_type_repository,
    EntityTypeBundleInfoInterface $entity_type_bundle_info,
    EntityTypeManagerInterface $entity_type_manager,
    StreamWrapperManagerInterface $stream_wrapper_manager,
    LanguageManagerInterface $language_manager,
    RemoteManagerInterface $remote_manager,
    EventDispatcherInterface $event_dispatcher,
    RequestServiceInterface $request_service,
    MessengerInterface $messenger,
    LoggerInterface $logger,
    StateInformationInterface $state_information
  ) {
    $this->jsonapiDocumentTopLevelNormalizer = $jsonapi_document_top_level_normalizer;
    $this->jsonapiDocumentTopLevelNormalizer->setSerializer($serializer);
    $this->resourceTypeRepository = $resource_type_repository;
    $this->bundleInfos = $entity_type_bundle_info->getAllBundleInfo();
    $this->entityDefinitions = $entity_type_manager->getDefinitions();
    $this->entityTypeManager = $entity_type_manager;
    $this->streamWrapperManager = $stream_wrapper_manager;
    $this->languageManager = $language_manager;
    $this->remoteManager = $remote_manager;
    $this->eventDispatcher = $event_dispatcher;
    $this->requestService = $request_service;
    $this->messenger = $messenger;
    $this->logger = $logger;
    $this->stateInformation = $state_information;
    $this->importedEntities = [];
  }

  /**
   * {@inheritdoc}
   */
  public function buildEntitiesOptions(array $json_data) {
    $options = [];
    foreach ($this->prepareData($json_data) as $data) {
      $this->addOptionFromJson($options, $data);
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function extractEntity(array $data) {
    // Format JSON as in
    // JsonApiDocumentTopLevelNormalizerTest::testDenormalize().
    $prepared_json = [
      'data' => [
        'type' => $data['type'],
        'attributes' => $data['attributes'],
      ],
    ];
    $parsed_type = explode('--', $data['type']);

    return $this->jsonapiDocumentTopLevelNormalizer->denormalize($prepared_json, NULL, 'api_json', [
      'resource_type' => $this->resourceTypeRepository->get(
        $parsed_type[0],
        $parsed_type[1]
      ),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function updateRelationships(ContentEntityInterface $entity, array $data) {
    if (isset($data['relationships'])) {
      $resource_type = $this->resourceTypeRepository->get(
        $entity->getEntityTypeId(),
        $entity->bundle()
      );
      // Reference fields.
      foreach ($data['relationships'] as $field_name => $field_data) {
        $field_name = $resource_type->getInternalName($field_name);
        $field = $entity->get($field_name);
        if ($this->relationshipHandleable($field)) {
          $field_values = [];

          // Check that the field has data.
          if ($field_data['data'] != NULL && isset($field_data['links']['related']['href'])) {
            $referenced_entities_response = $this->requestService->request($this->getHttpClient(), 'GET', $field_data['links']['related']['href']);
            $referenced_entities_json = Json::decode((string) $referenced_entities_response->getBody());

            // $referenced_entities_json['data'] can be null in the case of
            // missing/deleted referenced entities.
            if (!isset($referenced_entities_json['errors']) && !is_null($referenced_entities_json['data'])) {
              $referenced_entities_ids = $this->importEntityListData($referenced_entities_json['data']);

              $main_property = $field->getItemDefinition()->getMainPropertyName();

              // Remove the missing entities from the array to avoid key
              // mismatch.
              $prepared_data = [];
              foreach ($this->prepareData($field_data['data']) as $field_value_data) {
                if ($field_value_data['id'] !== 'missing') {
                  $prepared_data[] = $field_value_data;
                }
              }

              // Add field metadatas.
              foreach ($prepared_data as $key => $field_value_data) {
                // When dealing with taxonomy term entities which has a
                // hierarchy, there is a virtual entity for the root. So
                // $referenced_entities_ids[$key] may not exist.
                // See https://www.drupal.org/node/2976856.
                if (isset($referenced_entities_ids[$key])) {
                  $field_value = [
                    $main_property => $referenced_entities_ids[$key],
                  ];

                  if (isset($field_value_data['meta'])) {
                    $field_value += $field_value_data['meta'];
                  }

                  // Allow to alter the field value with an event.
                  $event = new RelationshipFieldValueEvent($field, $field_value);
                  $this->eventDispatcher->dispatch(RelationshipFieldValueEvent::EVENT_NAME, $event);
                  $field_values[] = $event->getFieldValue();
                }
              }
            }
          }
          $entity->set($field_name, $field_values);
        }
      }

      // Save the entity once all the references have been updated.
      $entity->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function handlePhysicalFiles(ContentEntityInterface $entity, array &$data) {
    if ($entity instanceof FileInterface) {
      // TODO: $resource_type->getPublicName().
      $remote_uri = $data['attributes']['uri']['value'];
      $remote_url = $data['attributes']['uri']['url'];
      $stream_wrapper = $this->streamWrapperManager->getViaUri($remote_uri);
      $directory_uri = $stream_wrapper->dirname($remote_uri);
      $log_variables = [
        '%url' => $remote_url,
        '%directory' => $directory_uri,
        '%id' => $entity->id(),
        '%uri' => $remote_uri,
      ];

      // Create the destination folder.
      if (file_prepare_directory($directory_uri, FILE_CREATE_DIRECTORY)) {
        try {
          $response = $this->requestService->request($this->getFileHttpClient(), 'GET', $remote_url);
          $file_content = (string) $response->getBody();
          $result = @file_put_contents($remote_uri, $file_content);
          if (!$result) {
            throw new \Exception('Error writing file to ' . $remote_uri);
          }
        }
        catch (ClientException $e) {
          $this->messenger->addWarning($this->t('Error importing file id %id. Missing file: %url', $log_variables));
          $this->logger->warning('Error importing file id %id. Missing file: %url', $log_variables);
        }
        catch (\Throwable $e) {
          $log_variables['@msg'] = $e->getMessage();
          $this->messenger->addError($this->t('Caught exception trying to import the file %url to %uri', $log_variables));
          $this->logger->error('Caught exception trying to import the file %url to %uri. Error message was @msg', $log_variables);
        }
      }
      else {
        $this->messenger->addError($this->t('Impossible to write in the directory %directory', $log_variables));
        $this->logger->error('Impossible to write in the directory %directory', $log_variables);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setRemote(RemoteInterface $remote) {
    $this->remote = $remote;
  }

  /**
   * {@inheritdoc}
   */
  public function importEntityListData(array $entity_list_data) {
    $imported_entity_ids = [];
    foreach ($this->prepareData($entity_list_data) as $entity_data) {
      $parsed_type = explode('--', $entity_data['type']);
      $entity_type_id = $parsed_type[0];
      $entity_storage = $this->entityTypeManager->getStorage($entity_type_id);
      $entity_keys = $entity_storage->getEntityType()->getKeys();

      $this->prepareEntityData($entity_data, $entity_keys);

      // TODO: $resource_type->getPublicName().
      $data_langcode = !empty($entity_keys['langcode']) ? $entity_data['attributes'][$entity_keys['langcode']] : LanguageInterface::LANGCODE_NOT_SPECIFIED;

      // Prepare entity label.
      if (isset($entity_keys['label'])) {
        // TODO: $resource_type->getPublicName().
        $entity_label = $entity_data['attributes'][$entity_keys['label']];
      }
      else {
        // Use the entity type if there is no label.
        $entity_label = $entity_type_id;
      }

      if ($data_langcode && !$this->dataLanguageExists($data_langcode, $entity_label)) {
        continue;
      }

      // Check if an entity already exists.
      // JSON:API no longer includes uuid in attributes so we're using id
      // instead. See https://www.drupal.org/node/2984247.
      $existing_entities = $entity_storage
        ->loadByProperties(['uuid' => $entity_data['id']]);

      // Here is the supposition that we are importing a list of content
      // entities. Currently this is ensured by the fact that it is not possible
      // to make a channel on config entities and on users. And that in the
      // relationshipHandleable() method we prevent handling config entities and
      // users relationships.
      /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
      $entity = $this->extractEntity($entity_data);

      // New entity.
      if (empty($existing_entities)) {
        $entity->save();
        $imported_entity_ids[] = $entity->id();
        // Prevent the entity of being reimported.
        $this->importedEntities[$entity->language()->getId()][$entity->uuid()] = $entity->uuid();
        $this->updateRelationships($entity, $entity_data);
        $this->handlePhysicalFiles($entity, $entity_data);
        // Change the entity "changed" time because it could have been altered
        // with relationship save by example.
        if (!empty($entity_data['attributes']['changed']) && method_exists($entity, 'setChangedTime')) {
          // TODO: $resource_type->getPublicName().
          // If the website is using backward compatible timestamps output.
          // @see https://www.drupal.org/node/2859657.
          if (is_numeric($entity_data['attributes']['changed'])) {
            // The value is casted in integer for
            // https://www.drupal.org/node/2837696.
            $entity->setChangedTime((int) $entity_data['attributes']['changed']);
          }
          elseif ($changed_datetime = \DateTime::createFromFormat(\DateTime::RFC3339, $entity_data['attributes']['changed'])) {
            $entity->setChangedTime($changed_datetime->getTimestamp());
          }
        }
        $entity->save();
      }
      // Update the existing entity.
      else {
        /** @var \Drupal\Core\Entity\ContentEntityInterface $existing_entity */
        $existing_entity = array_shift($existing_entities);
        $imported_entity_ids[] = $existing_entity->id();

        if (!isset($this->importedEntities[$data_langcode][$existing_entity->uuid()])) {
          // Prevent the entity translation of being reimported.
          $this->importedEntities[$data_langcode][$existing_entity->uuid()] = $existing_entity->uuid();
          $has_translation = $existing_entity->hasTranslation($data_langcode);
          // Update the existing translation.
          if ($has_translation) {
            $resource_type = $this->resourceTypeRepository->get(
              $entity->getEntityTypeId(),
              $entity->bundle()
            );
            $existing_translation = $existing_entity->getTranslation($data_langcode);
            foreach ($entity_data['attributes'] as $field_name => $value) {
              $field_name = $resource_type->getInternalName($field_name);
              $existing_translation->set(
                $field_name,
                $entity->get($field_name)->getValue()
              );
            }
            $existing_translation->save();
          }
          // Create the new translation.
          else {
            $translation = $entity->toArray();
            $existing_entity->addTranslation($data_langcode, $translation);
            $existing_entity->save();
            $existing_translation = $existing_entity->getTranslation($data_langcode);
          }
          $this->updateRelationships($existing_translation, $entity_data);
          $this->handlePhysicalFiles($existing_translation, $entity_data);
          // Change the entity "changed" time because it could have been altered
          // with relationship save by example.
          if (!empty($entity_data['attributes']['changed']) && method_exists($existing_translation, 'setChangedTime')) {
            // TODO: $resource_type->getPublicName().
            // If the website is using backward compatible timestamps output.
            // @see https://www.drupal.org/node/2859657.
            if (is_numeric($entity_data['attributes']['changed'])) {
              // The value is casted in integer for
              // https://www.drupal.org/node/2837696.
              $entity->setChangedTime((int) $entity_data['attributes']['changed']);
            }
            elseif ($changed_datetime = \DateTime::createFromFormat(\DateTime::RFC3339, $entity_data['attributes']['changed'])) {
              $entity->setChangedTime($changed_datetime->getTimestamp());
            }
          }
          $existing_translation->save();
        }
      }
    }
    return $imported_entity_ids;
  }

  /**
   * {@inheritdoc}
   */
  public function clearImportedEntities($langcode = '', $entity_uuid = '') {
    if (empty($langcode) && empty($entity_uuid)) {
      $this->importedEntities = [];
    }
    elseif (!empty($langcode) && empty(!$entity_uuid) && isset($this->importedEntities[$langcode][$entity_uuid])) {
      unset($this->importedEntities[$langcode][$entity_uuid]);
    }
    elseif (!empty($langcode) && isset($this->importedEntities[$langcode])) {
      $this->importedEntities[$langcode] = [];
    }
    elseif (!empty($entity_uuid)) {
      foreach ($this->importedEntities as $imported_entities_langcode => $imported_entities_uuids) {
        if (isset($this->importedEntities[$imported_entities_langcode][$entity_uuid])) {
          unset($this->importedEntities[$imported_entities_langcode][$entity_uuid]);
        }
      }
    }
  }

  /**
   * Helper function to add an option.
   *
   * @param array $options
   *   The array of options for the tableselect form type element.
   * @param array $data
   *   An array of data.
   * @param int $level
   *   The level of indentation.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \InvalidArgumentException
   */
  protected function addOptionFromJson(array &$options, array $data, $level = 0) {
    $indentation = '';
    for ($i = 1; $i <= $level; $i++) {
      $indentation .= '<div class="indentation">&nbsp;</div>';
    }

    $parsed_type = explode('--', $data['type']);
    $entity_type_id = $parsed_type[0];
    $bundle_id = $parsed_type[1];

    $entity_type = $this->entityTypeManager->getStorage($entity_type_id)->getEntityType();
    $entity_keys = $entity_type->getKeys();

    $label = '';
    // Some entity type may not have a label key and the label is calculated
    // using the label() method on the entity but at this step the entity is not
    // denormalized and also as we are not on the server website, we would not
    // have the data required to calculate the entity's label.
    if (isset($entity_keys['label']) && isset($data['attributes'][$entity_keys['label']])) {
      $label = new FormattableMarkup($indentation . '@label', [
        // TODO: $resource_type->getPublicName().
        '@label' => $data['attributes'][$entity_keys['label']],
      ]);
    }
    elseif (isset($entity_keys['id'])) {
      // TODO: $resource_type->getPublicName().
      // TODO: use $data['attributes'][$entity_keys['id']].
      $label = $data['id'];
    }

    $status_info = $this->stateInformation->getStatusInfo($data);

    $options[$data['id']] = [
      'label' => $label,
      'type' => $entity_type->getLabel(),
      'bundle' => $this->bundleInfos[$entity_type_id][$bundle_id]['label'],
      'language' => $this->getEntityLanguageLabel($data, $entity_keys),
      'status' => $status_info['label'],
      '#attributes' => [
        'class' => [
          $status_info['class'],
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function prepareData(array $data) {
    if ($this->isNumericArray($data)) {
      return $data;
    }
    else {
      return [$data];
    }
  }

  /**
   * Check if a array is numeric.
   *
   * @param array $array
   *   The array to check.
   *
   * @return bool
   *   TRUE if the array is numeric. FALSE in case of associative array.
   */
  protected function isNumericArray(array $array) {
    foreach ($array as $a => $b) {
      if (!is_int($a)) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * Check if a relationship is handleable.
   *
   * Filter on fields not targeting config entities or users.
   *
   * @param \Drupal\Core\Field\FieldItemListInterface $field
   *   The field item list.
   *
   * @return bool
   *   TRUE if the relationship is handleable.
   */
  protected function relationshipHandleable(FieldItemListInterface $field) {
    $relationship_handleable = FALSE;

    if ($field instanceof EntityReferenceFieldItemListInterface) {
      $settings = $field->getItemDefinition()->getSettings();

      // Entity reference and Entity reference revisions.
      if (isset($settings['target_type'])) {
        $relationship_handleable = !$this->isUserOrConfigEntity($settings['target_type']);
      }
      // Dynamic entity reference.
      elseif (isset($settings['entity_type_ids'])) {
        foreach ($settings['entity_type_ids'] as $entity_type_id) {
          $relationship_handleable = !$this->isUserOrConfigEntity($entity_type_id);
          if (!$relationship_handleable) {
            break;
          }
        }
      }
    }

    return $relationship_handleable;
  }

  /**
   * Helper function to get the language from an extracted entity.
   *
   * We can't use $entity->language() because if the entity is in a language not
   * enabled, it is the site default language that is returned.
   *
   * @param array $data
   *   The data from the JSON API payload.
   * @param array $entity_keys
   *   The entity keys from the entity definition.
   *
   * @return string
   *   The language of the entity.
   */
  protected function getEntityLanguageLabel(array $data, array $entity_keys) {
    if (!isset($entity_keys['langcode']) || empty($entity_keys['langcode'])) {
      return $this->t('Untranslatable entity');
    }

    // TODO: $resource_type->getPublicName().
    $langcode = $data['attributes'][$entity_keys['langcode']];
    $language = $this->languageManager->getLanguage($langcode);
    // Check if the entity is in an enabled language.
    if (is_null($language)) {
      $language_list = LanguageManager::getStandardLanguageList();
      if (isset($language_list[$langcode])) {
        $entity_language = $language_list[$langcode][0] . ' ' . $this->t('(not enabled)', [], ['context' => 'language']);
      }
      else {
        $entity_language = $this->t('Entity in an unsupported language.');
      }
    }
    else {
      $entity_language = $language->getName();
    }

    return $entity_language;
  }

  /**
   * Helper function to get the File Http Client.
   *
   * @return \GuzzleHttp\Client
   *   A HTTP client to retrieve files.
   */
  protected function getFileHttpClient() {
    if (!$this->fileHttpClient) {
      $this->fileHttpClient = $this->remoteManager->prepareClient($this->remote);
    }

    return $this->fileHttpClient;
  }

  /**
   * Helper function to get the Http Client.
   *
   * @return \GuzzleHttp\Client
   *   A HTTP client to request JSON:API endpoints.
   */
  protected function getHttpClient() {
    if (!$this->httpClient) {
      $this->httpClient = $this->remoteManager->prepareJsonApiClient($this->remote);
    }

    return $this->httpClient;
  }

  /**
   * Prepare the data array before extracting the entity.
   *
   * Used to remove some data.
   *
   * @param array $data
   *   An array of data.
   * @param array $entity_keys
   *   An array of entity keys.
   */
  protected function prepareEntityData(array &$data, array $entity_keys) {
    $parsed_type = explode('--', $data['type']);

    $resource_type = $this->resourceTypeRepository->get(
      $parsed_type[0],
      $parsed_type[1]
    );

    // Removes some ids.
    unset($data['attributes'][$resource_type->getPublicName($entity_keys['id'])]);
    if (isset($entity_keys['revision']) && !empty($entity_keys['revision'])) {
      unset($data['attributes'][$resource_type->getPublicName($entity_keys['revision'])]);
    }

    // UUID is no longer included as attribute.
    $data['attributes'][$resource_type->getPublicName($entity_keys['uuid'])] = $data['id'];

    // Remove the default_langcode boolean to be able to import content not
    // necessarily in the default language.
    unset($data['attributes'][$resource_type->getPublicName($entity_keys['default_langcode'])]);

    // To avoid side effects and as currently JSON:API send null for the path
    // we remove the path attribute.
    // TODO: $resource_type->getPublicName().
    if (isset($data['attributes']['path'])) {
      unset($data['attributes']['path']);
    }
  }

  /**
   * Check if we try to import an entity in a disabled language.
   *
   * @param string $langcode
   *   The langcode of the language to check.
   * @param string $entity_label
   *   The entity label.
   *
   * @return bool
   *   FALSE if the data is not in an enabled language.
   */
  protected function dataLanguageExists($langcode, $entity_label) {
    if (is_null($this->languageManager->getLanguage($langcode))) {
      $this->messenger->addError($this->t('Trying to import an entity (%entity_label) in a disabled language.', [
        '%entity_label' => $entity_label,
      ]));
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Helper function to check if an entity type id is a user or a config entity.
   *
   * @param string $entity_type_id
   *   The entity type id.
   *
   * @return bool
   *   TRUE if the entity type is user or a config entity. FALSE otherwise.
   */
  protected function isUserOrConfigEntity($entity_type_id) {
    if ($entity_type_id == 'user') {
      return TRUE;
    }
    elseif ($this->entityDefinitions[$entity_type_id]->getGroup() == 'configuration') {
      return TRUE;
    }

    return FALSE;
  }

}
